@@ -102,7 +102,7 @@ This method is called when L<LISM> commit the update of storage.
 
 sub commit
 {
-    die("must be overridden");
+    return 0;
 }
 
 =pod
@@ -115,7 +115,7 @@ This method is called when L<LISM> rollback the update of storage.
 
 sub rollback
 {
-    die("must be overridden");
+    return 0;
 }
 
 =pod
@@ -187,6 +187,10 @@ sub search
         return (LDAP_FILTER_ERROR, ());
     }
 
+    if (!$sizeLim) {
+        $sizeLim = -1;
+    }
+
     my $retry = 0;
     while (1) {
         if (!$self->_getConnect()) {
@@ -302,7 +306,6 @@ sub compare
         }
 
         # compare the value
-        $entry = encode('utf8', $entry);
         if ($entry =~ /^$attr: $val$/m) {
             $rc = LDAP_COMPARE_TRUE;
         } else {
@@ -368,7 +371,8 @@ sub add
 
     # check rdn's value
     my ($rdn, $rdn_val) = ($dn =~ /^([^=]+)=([^,]+),/);
-    if ($entryStr !~ /^$rdn: $rdn_val$/mi) {
+    $rdn_val =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
+    if ($entryStr !~ /^$rdn: $rdn_val *$/mi) {
         return LDAP_NAMING_VIOLATION;
     }
 
@@ -503,6 +507,10 @@ sub log
     my $conf = $self->{_config};
     my %p = @_;
 
+    if (Encode::is_utf8($p{'message'})) {
+        $p{'message'} = encode('utf8', $p{'message'});
+    }
+
     openlog('LISM', 'pid', 'local4');
     setlogmask(Sys::Syslog::LOG_UPTO($conf->{numloglevel}{$conf->{sysloglevel}}));
     syslog($conf->{numloglevel}{$p{'level'}}, $p{'message'});
@@ -680,7 +688,7 @@ sub _checkConfig
 
     # multibyte character code
     if (defined($conf->{mbcode})) {
-        if ($conf->{mbcode}[0] !~ /^(utf8|euc-jp|shiftjis)/) {
+        if ($conf->{mbcode}[0] !~ /^(utf8|euc-jp|shiftjis|cp932)/) {
             $self->log(level => 'alert', message => "character code is invalid");
             return 1;
         }
@@ -688,6 +696,19 @@ sub _checkConfig
         $conf->{mbcode}[0] = 'utf8';
     }
 
+    # break character in value
+    if (defined($conf->{break})) {
+        if ($conf->{break}[0] eq 'CR') {
+            $conf->{breakchar} = "\r";
+        } elsif ($conf->{break}[0] eq 'CRLF') {
+            $conf->{breakchar} = "\r\n";
+        } else {
+            $conf->{breakchar} = "\n";
+        }
+    } else {
+        $conf->{breakchar} = "\n";
+    }
+
     # connection
     if (!defined($conf->{connection}->{retry})
         || $conf->{connection}->{retry}[0] !~ /[0-9]*/) {
@@ -709,6 +730,7 @@ sub _checkConfig
                 $self->log(level => 'alert', message => "$oname object container doesn't exist");
                 return 1;
             }
+            $self->{object}{$oname}->{name} = $oname;
             $self->{object}{$oname}->{conf} = $oconf;
 
             # set container
@@ -742,11 +764,6 @@ sub _checkConfig
                 next;
             }
 
-            # check id definition
-            if (!defined($oconf->{id}) || !defined($oconf->{id}[0]->{column})) {
-                $self->log(level => 'alert', message => "Id definition is invalid");
-            }
-
             foreach my $attr (keys(%{$oconf->{attr}})) {
                 # the attribute's name must be lowercase
                 if ($attr =~ /[A-Z]/) {
@@ -766,6 +783,16 @@ sub _checkConfig
                 $self->log(level => 'alert', message => "rdn of $oname object is invalid");
                 return 1;
             }
+
+            if (defined($oconf->{strginfo})) {
+                foreach my $si (@{$oconf->{strginfo}}) {
+                    # check type of value
+                    if (defined($si->{value}) && !ref($si->{value}[0])) {
+                        $self->log(level => 'alert', message => "type of strginfo value doesn't exist");
+                        return 1;
+                    }
+                }
+            }
         }
 
         # check container's link
@@ -947,12 +974,9 @@ sub _baseSearch
     }
 
     my ($rdn, $pdn) = ($base =~ /^([^,]+),(.*)$/);
-
-    my $filterStr;
+    my $filterStr = "(".encode('utf8', $rdn).")";
     if ($filter) {
-        $filterStr = "(&($rdn)".$filter->as_string.")";
-    } else {
-        $filterStr = "($rdn)";
+        $filterStr = "(&".$filterStr.$filter->as_string.")";
     }
     my $basefilter = Net::LDAP::Filter->new($filterStr);
 
@@ -1101,6 +1125,27 @@ sub _getStaticValue
     return $value;
 }
 
+sub _funcParse
+{
+    my $self = shift;
+    my ($str, $dn, $entryStr) = @_;
+
+    my @funcs = ($str =~ /%{([^}]*)}/g);
+    foreach my $func (@funcs) {
+        my $value;
+
+        eval "\$value = $func";
+        if ($@) {
+            $self->log(level => 'error', message => "function $func failed: $@");
+        }
+
+        $func =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
+        $str =~ s/%{$func}/$value/;
+    }
+
+    return $str;
+}
+
 sub _pwdFormat
 {
     my $self = shift;
