@@ -10,7 +10,7 @@ use Sys::Syslog;
 use Sys::Syslog qw(:macros);
 use Data::Dumper;
 
-our $rawattrs = '^(jpeg|photo|objectSid|objectGUID|.*;binary)$';
+our $rawattrs = '^(jpegphoto|photo|objectSid|objectGUID|.*;binary)$';
 
 =head1 NAME
 
@@ -35,10 +35,11 @@ sub init
 
     $self->SUPER::init();
 
+    if (!Encode::is_utf8($conf->{uri}[0])) {
+        $conf->{uri}[0] = decode('utf8', $conf->{uri}[0]);
+    }
+
     if (!defined($conf->{nc})) {
-        if (utf8::is_utf8($conf->{uri}[0])) {
-            $conf->{uri}[0] = encode('utf8', $conf->{uri}[0]);
-        }
         ($conf->{nc}) = ($conf->{uri}[0] =~ /^ldaps?:\/\/[^\/]+\/(.+)$/i);
     }
 
@@ -86,7 +87,6 @@ sub bind
     my $conf = $self->{_config};
 
     $binddn =~ s/$self->{suffix}$/$conf->{nc}/i;
-    $binddn = decode('utf8', $binddn);
 
     if ($self->_getConnect()) {
         return LDAP_SERVER_DOWN;
@@ -110,95 +110,8 @@ Search LDAP information.
 sub search
 {
     my $self = shift;
-    my($base, $scope, $deref, $sizeLim, $timeLim, $filterStr, $attrOnly, @attrs) = @_;
-    my $conf = $self->{_config};
-    my @match_entries = ();
-    my $rc = LDAP_SUCCESS;
 
-    my $filter = Net::LDAP::Filter->new($filterStr);
-    if (!defined($filter)) {
-        return (LDAP_FILTER_ERROR, ());
-    }
-
-    # get entry of data container
-    if ($base =~ /^$self->{suffix}$/) {
-        if ($scope != 1) {
-            my $entry = $self->{contentrystr};
-            if ($self->parseFilter($filter, $entry)) {
-                push (@match_entries, $entry);
-                $sizeLim--;
-            }
-        }
-    }
-    $sizeLim = $sizeLim - @match_entries;
-
-    if ($self->_getConnect()) {
-        return LDAP_SERVER_DOWN;
-    }
-
-    # Attribute mapping
-    foreach my $ldapmap (@{$conf->{ldapmap}}) {
-        if ($ldapmap->{type} =~ /^objectclass$/i) {
-            $filterStr =~ s/objectClass=$ldapmap->{local}$/objectClass=$ldapmap->{foreign}/mi;
-        } else {
-            $filterStr =~ s/\($ldapmap->{local}=/\($ldapmap->{foreign}=/mi;
-        }
-
-        if (Encode::is_utf8($filterStr)) {
-            Encode::_utf8_off($filterStr);
-        }
-    }
-
-    $base =~ s/$self->{suffix}$/$conf->{nc}/i;
-    $base = decode('utf8', $base);
-    $filterStr =~ s/$self->{suffix}(\)*)/$conf->{nc}$1/gi;
-    my $msg = $self->{ldap}->search(base => $base, scope => $scope, deref => $deref, sizeLimit => $sizeLim, timeLimit => $timeLim, filter => $filterStr);
-
-    if (!$msg->code) {
-        for (my $i = 0; $i < $msg->count; $i++) {
-            my $entry = $msg->entry($i);
-            my $dn = $entry->dn;
-
-            $dn =~ s/$conf->{nc}$/$self->{suffix}/i;
-            my $entryStr = "dn: $dn\n";
-            if ($dn =~ /^$self->{suffix}$/i) {
-                next;
-            } else {
-                foreach my $attr ($entry->attributes) {
-                    foreach my $value ($entry->get_value($attr)) {
-                        if ($attr =~ /$rawattrs/i) {
-                            $value = encode_base64($value);
-                            $entryStr = $entryStr.$attr.":: $value";
-                        } else {
-                            $value =~ s/$conf->{nc}$/$self->{suffix}/i;
-                            $entryStr = $entryStr.$attr.": $value\n";
-                        }
-                    }
-                }
-            }
-
-            if (!$self->_checkEntry($entryStr)) {
-                next;
-            }
-
-            # Attribute mapping
-            foreach my $ldapmap (@{$conf->{ldapmap}}) {
-                if ($ldapmap->{type} =~ /^objectclass$/i) {
-                    $entryStr =~ s/^objectClass: $ldapmap->{foreign}$/objectClass: $ldapmap->{local}/mi;
-                } else {
-                    $entryStr =~ s/^$ldapmap->{foreign}:/$ldapmap->{local}:/mi;
-                }
-            }
-
-            $entryStr = decode('utf8', $entryStr);
-
-            push(@match_entries, $entryStr);
-        }
-    }
-
-    $self->_freeConnect($msg);
-
-    return ($msg->code , @match_entries);
+    return $self->_do_search(undef, @_);
 }
 
 =pod
@@ -216,7 +129,6 @@ sub compare
     my $conf = $self->{_config};
 
     $dn =~ s/$self->{suffix}$/$conf->{nc}/i;
-    $dn = decode('utf8', $dn);
 
     my ($key, $val) = split(/=/, $avaStr);
 
@@ -259,7 +171,6 @@ sub modify
     my $conf = $self->{_config};
 
     $dn =~ s/$self->{suffix}$/$conf->{nc}/i;
-    $dn = decode('utf8', $dn);
 
     my @changes;
     while ( @list > 0) {
@@ -297,7 +208,8 @@ sub modify
         for (my $i = 0; $i < @values; $i++) {
             $values[$i] =~ s/$self->{suffix}/$conf->{nc}/i;
             if ($values[$i]) {
-                $values[$i] = decode('utf8', $values[$i]);
+                # replace carriage return to linefeed
+                $values[$i] =~ s/\r/$conf->{breakchar}/g;
             }
         }
 
@@ -334,8 +246,6 @@ sub add
     my $conf = $self->{_config};
 
     $dn =~ s/$self->{suffix}$/$conf->{nc}/i;
-    $dn = decode('utf8', $dn);
-    $entryStr = decode('utf8', $entryStr);
 
     my %attrs;
     my @info = split(/\n/, $entryStr);
@@ -363,6 +273,10 @@ sub add
         }
 
         $val =~ s/$self->{suffix}/$conf->{nc}/i;
+
+        # replace carriage return to linefeed
+        $val =~ s/\r/$conf->{breakchar}/g;
+
         push(@{$attrs{$key}}, $val);
     }
 
@@ -393,8 +307,6 @@ sub modrdn
     my $conf = $self->{_config};
 
     $dn =~ s/$self->{suffix}$/$conf->{nc}/i;
-    $dn = decode('utf8', $dn);
-    $newrdn = decode('utf8', $newrdn);
 
     if ($self->_getConnect()) {
         return LDAP_SERVER_DOWN;
@@ -422,7 +334,6 @@ sub delete
     my $conf = $self->{_config};
 
     $dn =~ s/$self->{suffix}$/$conf->{nc}/i;
-    $dn = decode('utf8', $dn);
 
     if ($self->_getConnect()) {
         return LDAP_SERVER_DOWN;
@@ -506,6 +417,10 @@ sub _freeConnect
     my $self =shift;
     my ($msg) = @_;
 
+    if (!$msg) {
+        return -1;
+    }
+
     if ($msg->code == LDAP_SERVER_DOWN || $msg->code == -1) {
         $self->{ldap}->unbind();
         $self->{bind}->unbind();
@@ -513,6 +428,8 @@ sub _freeConnect
         undef($self->{ldap});
         undef($self->{bind});
     }
+
+    return 0;
 }
 
 sub _checkConfig
@@ -534,6 +451,108 @@ sub _checkConfig
     return $rc;
 }
 
+sub _do_search
+{
+    my $self = shift;
+    my($control, $base, $scope, $deref, $sizeLim, $timeLim, $filterStr, $attrOnly, @attrs) = @_;
+    my $conf = $self->{_config};
+    my @match_entries = ();
+    my $rc = LDAP_SUCCESS;
+
+    my $filter = Net::LDAP::Filter->new($filterStr);
+    if (!defined($filter)) {
+        return (LDAP_FILTER_ERROR, ());
+    }
+
+    # get entry of data container
+    if ($base =~ /^$self->{suffix}$/) {
+        if ($scope != 1) {
+            my $entry = $self->{contentrystr};
+            if ($self->parseFilter($filter, $entry)) {
+                push (@match_entries, $entry);
+                $sizeLim--;
+            }
+        }
+    }
+    $sizeLim = $sizeLim - @match_entries;
+
+    if ($self->_getConnect()) {
+        return LDAP_SERVER_DOWN;
+    }
+
+    $filterStr = decode('utf8', $filterStr);
+
+    # Attribute mapping
+    foreach my $ldapmap (@{$conf->{ldapmap}}) {
+        if ($ldapmap->{type} =~ /^objectclass$/i) {
+            $filterStr =~ s/objectClass=$ldapmap->{local}$/objectClass=$ldapmap->{foreign}/mi;
+        } else {
+            $filterStr =~ s/\($ldapmap->{local}=/\($ldapmap->{foreign}=/mi;
+        }
+    }
+
+    $base =~ s/$self->{suffix}$/$conf->{nc}/i;
+    $filterStr =~ s/$self->{suffix}(\)*)/$conf->{nc}$1/gi;
+    $filterStr = encode('utf8', $filterStr);
+
+    my $msg;
+    if ($control) {
+        $msg = $self->{ldap}->search(base => $base, scope => $scope, deref => $deref, sizeLimit => $sizeLim, timeLimit => $timeLim, filter => $filterStr, control => $control);
+    } else {
+        $msg = $self->{ldap}->search(base => $base, scope => $scope, deref => $deref, sizeLimit => $sizeLim, timeLimit => $timeLim, filter => $filterStr);
+    }
+
+    if (!$msg->code) {
+        for (my $i = 0; $i < $msg->count; $i++) {
+            my $entry = $msg->entry($i);
+            my $dn = decode('utf8', $entry->dn);
+
+            $dn =~ s/$conf->{nc}$/$self->{suffix}/i;
+            my $entryStr = "dn: $dn\n";
+            if ($dn =~ /^$self->{suffix}$/i) {
+                next;
+            } else {
+                foreach my $attr ($entry->attributes) {
+                    foreach my $value ($entry->get_value($attr)) {
+                        if ($attr =~ /$rawattrs/i) {
+                            $value = encode_base64($value, '');
+                            $entryStr = $entryStr.$attr.":: $value\n";
+                        } else {
+                            if ($value =~ /\n/) {
+                                $value = encode_base64($value, '');
+                                $entryStr = $entryStr.$attr.":: $value\n";
+                            } else {
+                                $value = decode('utf8', $value);
+                                $value =~ s/$conf->{nc}$/$self->{suffix}/i;
+                                $entryStr = $entryStr.$attr.": $value\n";
+                            }
+                        }
+                    }
+                }
+            }
+
+            if (!$self->_checkEntry($entryStr)) {
+                next;
+            }
+
+            # Attribute mapping
+            foreach my $ldapmap (@{$conf->{ldapmap}}) {
+                if ($ldapmap->{type} =~ /^objectclass$/i) {
+                    $entryStr =~ s/^objectClass: $ldapmap->{foreign}$/objectClass: $ldapmap->{local}/mi;
+                } else {
+                    $entryStr =~ s/^$ldapmap->{foreign}:/$ldapmap->{local}:/mi;
+                }
+            }
+
+            push(@match_entries, $entryStr);
+        }
+    }
+
+    $self->_freeConnect($msg);
+
+    return ($msg->code , @match_entries);
+}
+
 sub _checkEntry
 {
     my $self = shift;
