@@ -5,11 +5,14 @@ use base qw(LISM::Storage);
 use Net::LDAP::Filter;
 use Net::LDAP::Constant qw(:all);
 use DBI;
+use MIME::Base64;
 use Encode;
 use Sys::Syslog;
 use Sys::Syslog qw(:macros);
 use Data::Dumper;
 
+our $rawattrs = '^(jpegphoto|photo|.*;binary)$';
+
 =head1 NAME
 
 LISM::Storage::SQL - SQL storage for LISM
@@ -229,7 +232,7 @@ sub _objSearch
 
     DO: {
         # get data of the entries
-        my $sql = "select distinct $oconf->{table}[0].$oconf->{id}[0]->{column}[0], $oconf->{table}[0].$oconf->{attr}{$oconf->{rdn}[0]}->{column}[0]";
+        my $sql = "select distinct $oconf->{table}[0].$oconf->{id}[0]->{column}[0]";
         foreach my $attr ('objectclass', keys %{$oconf->{attr}}) {
             if (!defined($oconf->{attr}{$attr}) || !defined($oconf->{attr}{$attr}->{column})) {
                 next;
@@ -261,10 +264,17 @@ sub _objSearch
             }
         }
 
+        foreach my $strginfo (@{$oconf->{strginfo}}) {
+            if (defined($strginfo->{selwhere})) {
+                $where = "$where and $strginfo->{selwhere}[0]";
+            }
+        }
+
         $sql = "$sql from $from where $where";
 
-        # multibyte value
-        $sql = encode($conf->{mbcode}[0], decode('utf8', $sql));
+        # encode value
+        $sql = encode($conf->{mbcode}[0], $sql);
+
         my $sth = $self->{db}->prepare($sql);
         if (!$sth->execute) {
             $self->log(level => 'error', message => "Searching by \"$sql\" failed: ".$sth->errstr);
@@ -288,10 +298,6 @@ sub _objSearch
                 next;
             }
 
-            my $rdn_val = shift(@data);
-            $rdn_val =~ s/ *$//;
-            my $rdn = "$oconf->{rdn}[0]=$rdn_val";
-
             my $j = 0;
             foreach my $attr ('objectclass', keys %{$oconf->{attr}}) {
                 if ($attr =~ /^objectclass$/i) {
@@ -301,15 +307,24 @@ sub _objSearch
                 } elsif (!defined($oconf->{attr}->{$attr})) {
                     next;
                 } elsif (defined($oconf->{attr}{$attr}->{column})) {
-                    if ($data[$j]) {
-                        my $value = $data[$j];
-                        $value =~ s/ *$//;
-                        $entry = $entry."$attr: $value\n";
+                    my $value = $data[$j];
+                    $value =~ s/ *$//;
+                    if ($value) {
+                        if ($attr =~ /$rawattrs/i) {
+                            $value = encode_base64($value, '');
+                            $entry = $entry.$attr.":: $value\n";
+                        } elsif ($value =~ /\n/) {
+                            Encode::from_to($value, $conf->{mbcode}[0], 'utf8');
+                            $value = encode_base64($value, '');
+                            $entry = $entry.$attr.":: $value\n";
+                        } else {
+                            $entry = $entry."$attr: $value\n";
+                        }
                     }
                     $j++;
                 } else {
                     # get the values from the attribute's table
-                    my $values = $self->_getAttrValues($oconf, $attr, $key);
+                    my $values = $self->_getAttrValues($oconf, $key, $pkey, $attr);
                     if (!defined($values)) {
                         $rc = LDAP_OPERATIONS_ERROR;
                         @match_entries = ();
@@ -325,20 +340,16 @@ sub _objSearch
             # multibyte value
             if ($conf->{dsn}[0] =~ /^DBI:Oracle/i && Encode::is_utf8($entry)) {
                 # add-hoc method for Oracle
-                if (Encode::is_utf8($rdn)) {
-                    Encode::_utf8_off($rdn);
-                }
-                if (Encode::is_utf8($entry)) {
-                    Encode::_utf8_off($entry);
-                }
+                Encode::_utf8_off($entry);
             }
 
-            Encode::from_to($rdn, $conf->{mbcode}[0], 'utf8');
-            $rdn = decode('utf8', $rdn);
             Encode::from_to($entry, $conf->{mbcode}[0], 'utf8');
             $entry = decode('utf8', $entry);
 
+            my ($rdn_val) = ($entry =~ /^$oconf->{rdn}[0]: (.*)$/mi);
+            my $rdn = "$oconf->{rdn}[0]=$rdn_val";
             $entry = "dn: $rdn,$suffix\n$entry";
+
             push(@match_entries, $self->_pwdFormat($entry));
             push(@match_keys, $key);
         }
@@ -385,13 +396,11 @@ sub _objModify
             $rc = LDAP_NO_SUCH_OBJECT;
             last DO;
         }
-        $entry = encode('utf8', $entry);
 
         while ( @list > 0 ) {
             my $action = shift @list;
             my $attr    = lc(shift @list);
             my @values;
-            my $value;
             my $sql;
             my $sth;
 
@@ -405,14 +414,24 @@ sub _objModify
 
             # can't modify the attribute for rdn
             if ($attr eq $oconf->{rdn}[0]) {
-                if ($action ne "REPLACE" || @values != 1 || !($entry =~ /^$attr: $values[0]$/mi)) {
+                if ($action ne "REPLACE" || @values != 1) {
                     $rc = LDAP_CONSTRAINT_VIOLATION;
                     last DO;
+                } else {
+                    my $value = $values[0];
+                    $value =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
+                    if ($entry !~ /^$attr: $value$/mi) {
+                        $rc = LDAP_CONSTRAINT_VIOLATION;
+                        last DO;
+                    }
                 }
             }
 
             for (my $i = 0; $i < @values; $i++) {
                 if ($values[$i]) {
+                    # replace carriage return to linefeed
+                    $values[$i] =~ s/\r/$conf->{breakchar}/g;
+
                     # escape special character
                     $values[$i] =~ s/'/''/g;
                     $values[$i] =~ s/\\/\\\\/g;
@@ -422,6 +441,7 @@ sub _objModify
             if($action eq "ADD") {
                 # check whether the value already exists
                 my $vals_str = "(".join('|', @values).")";
+                $vals_str =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
                 if ($entry =~ /^$attr: $vals_str *$/mi) {
                     $rc = LDAP_TYPE_OR_VALUE_EXISTS;
                     last DO;
@@ -436,7 +456,7 @@ sub _objModify
                     $sql = "update $oconf->{table}[0] set $oconf->{attr}{$attr}->{column}[0] = \'$values[0]\' where $oconf->{id}[0]->{column}[0] = \'$key\'";
 
                     # multibyte value
-                    $sql = encode($conf->{mbcode}[0], decode('utf8', $sql));
+                    $sql = encode($conf->{mbcode}[0], $sql);
 
                     $sth = $self->{db}->prepare($sql);
                     if (!$sth->execute) {
@@ -448,8 +468,8 @@ sub _objModify
                 }
 
                 if (defined($oconf->{attr}{$attr}->{addproc})) {
-                    foreach $value (@values) {
-                        if ($self->_addAttrValues($oconf, $key, $attr, $value)) {
+                    for (my $i = 0; $i < @values; $i++) {
+                        if ($self->_addAttrValues($oconf, $key, $pkey, $attr, $values[$i], $dn)) {
                             $rc = LDAP_OPERATIONS_ERROR;
                             last DO;
                         }
@@ -457,23 +477,20 @@ sub _objModify
                 }
             } elsif( $action eq "DELETE" ) {
                 # check whether the value exists
-                foreach $value (@values) {
-                    if ($value && !($entry =~ /^$attr: $value *$/mi)) {
+                for (my $i = 0; $i < @values; $i++) {
+                    my $value = $values[$i];
+                    $value =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
+                    if ($value && $entry !~ /^$attr: $value *$/mi) {
                         $rc = LDAP_NO_SUCH_ATTRIBUTE;
                         last DO;
                     }
                 }
 
                 if (defined($oconf->{attr}{$attr}->{column})) {
-                    if (@values && (@values > 1 || $values[0] && !($entry =~ /^$attr: $values[0] *$/mi))) {
-                        $rc = LDAP_NO_SUCH_ATTRIBUTE;
-                        last DO;
-                    }
                     $sql = "update $oconf->{table}[0] set $oconf->{attr}{$attr}->{column}[0] = \'\' where $oconf->{id}[0]->{column}[0] = \'$key\'";
 
                     # multibyte value
-                    $sql = encode($conf->{mbcode}[0], decode('utf8', $sql));
-
+                    $sql = encode($conf->{mbcode}[0], $sql);
 
                     $sth = $self->{db}->prepare($sql);
                     if (!$sth->execute) {
@@ -487,10 +504,15 @@ sub _objModify
                 if (defined($oconf->{attr}{$attr}->{delproc})) {
                     if (!@values || !$values[0]) {
                         @values = ($entry =~ /^$attr: (.*)$/mgi);
+                        for (my $i = 0; $i < @values; $i++) {
+                            # escape special character
+                            $values[$i] =~ s/'/''/g;
+                            $values[$i] =~ s/\\/\\\\/g;
+                        }
                     }
 
-                    foreach $value (@values) {
-                        if ($self->_delAttrValues($oconf, $key, $attr, $value)) {
+                    foreach my $value (@values) {
+                        if ($self->_delAttrValues($oconf, $key, $pkey, $attr, $value)) {
                             $rc = LDAP_OPERATIONS_ERROR;
                             last DO;
                         }
@@ -506,7 +528,7 @@ sub _objModify
                     $sql = "update $oconf->{table}[0] set $oconf->{attr}{$attr}->{column}[0] = \'$values[0]\' where $oconf->{id}[0]->{column}[0] = \'$key\'";
 
                     # multibyte value
-                    $sql = encode($conf->{mbcode}[0], decode('utf8', $sql));
+                    $sql = encode($conf->{mbcode}[0], $sql);
 
                     $sth = $self->{db}->prepare($sql);
                     if (!$sth->execute) {
@@ -524,7 +546,9 @@ sub _objModify
                     # delete the old values which exists
                     foreach my $old_val (@old_vals) {
                         my $match = 0;
-                        foreach $value (@values) {
+                        for (my $i = 0; $i < @values; $i++) {
+                            my $value = $values[$i];
+                            $value =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
                             if ($old_val =~ /^$value$/i) {
                                 $match = 1;
                                 next;
@@ -534,18 +558,26 @@ sub _objModify
                             push(@existing_vals, $old_val);
                             next;
                         }
-                        if ($self->_delAttrValues($oconf, $key, $attr, $old_val)) {
+
+                        # escape special character
+                        $old_val =~ s/'/''/g;
+                        $old_val =~ s/\\/\\\\/g;
+
+                        if ($self->_delAttrValues($oconf, $key, $pkey, $attr, $old_val)) {
                             $rc = LDAP_OPERATIONS_ERROR;
                             last DO;
                         }
                     }
 
                     # add the new values doesn't exist
-                    foreach $value (@values) {
+                    for (my $i = 0; $i < @values; $i++) {
+                        my $value = $values[$i];
+                        $value =~ s/([.*+?\[\]()|\^\$\\])/\\$1/g;
                         if (grep(/^$value$/i, @existing_vals)) {
                             next;
                         }
-                        if ($self->_addAttrValues($oconf, $key, $attr, $value)) {
+
+                        if ($self->_addAttrValues($oconf, $key, $pkey, $attr, $values[$i], $dn)) {
                             $rc = LDAP_OPERATIONS_ERROR;
                             last DO;
                         }
@@ -614,6 +646,9 @@ sub _objAdd
                 next;
             }
 
+            # repace carriage return to linefeed
+            $value =~ s/\r/$conf->{breakchar}/g;
+
             # escape special character
             $value =~ s/'/''/g;
             $value =~ s/\\/\\\\/g;
@@ -646,7 +681,7 @@ sub _objAdd
         my $sql = "insert into $oconf->{table}[0](".join(', ', @cols).") values(".join(', ', @values).")";
 
         # multibyte value
-        $sql = encode($conf->{mbcode}[0], decode('utf8', $sql));
+        $sql = encode($conf->{mbcode}[0], $sql);
 
         my $sth = $self->{db}->prepare("$sql");
         if (!$sth->execute) {
@@ -658,7 +693,7 @@ sub _objAdd
 
         # get the added object's id from the table
         my ($rdn, $pdn) = ($dn=~ /^([^,]+),(.*)$/);
-        my $filter = Net::LDAP::Filter->new("($rdn)");
+        my $filter = Net::LDAP::Filter->new("(".encode('utf8', $rdn).")");
         my $keys;
 
         ($rc, $keys) = $self->_objSearch($obj, '', $pdn, -1, $filter);
@@ -684,9 +719,10 @@ sub _objAdd
 	        $sql = $addproc;
                 $sql =~ s/\%o/$key/g;
                 $sql =~ s/\%v/$si_values[$i]/g;
+                $sql = $self->_funcParse($sql, $dn, $entryStr);
 
                 # multibyte value
-                $sql = encode($conf->{mbcode}[0], decode('utf8', $sql));
+                $sql = encode($conf->{mbcode}[0], $sql);
 
                 $sth = $self->{db}->prepare("$sql");
                 if (!$sth->execute) {
@@ -701,7 +737,7 @@ sub _objAdd
         # add the values in the attribute's table
         foreach my $attr (keys %attrs) {
             foreach my $value (@{$attrs{$attr}}) {
-                if ($rc = $self->_addAttrValues($oconf, $key, $attr, $value)) {
+                if ($rc = $self->_addAttrValues($oconf, $key, $pkey, $attr, $value, $dn, $entryStr)) {
                     $self->log(level => 'error', message => "Adding values of $attr failed");
                     $rc = LDAP_OPERATIONS_ERROR;
                     last DO;
@@ -714,6 +750,7 @@ sub _objAdd
             $sql = $oconf->{container}[0]->{addproc}[0];
             $sql =~ s/\%o/$key/g;
             $sql =~ s/\%c/$pkey/g;
+            $sql = $self->_funcParse($sql, $dn, $entryStr);
             $sth = $self->{db}->prepare("$sql");
             if (!$sth->execute) {
                 $self->log(level => 'error', message => "Adding link of container by \"$sql\" failed: ".$sth->errstr);
@@ -773,7 +810,7 @@ sub _objDelete
         # delete the values from the attribute's table
         foreach my $attr (keys %{$oconf->{attr}}) {
             if (defined($oconf->{attr}{$attr}) && defined($oconf->{attr}{$attr}->{delproc})) {
-                if ($rc = $self->_delAttrValues($oconf, $key, $attr, '')) {
+                if ($rc = $self->_delAttrValues($oconf, $key, $pkey, $attr, '')) {
                     $rc = LDAP_OPERATIONS_ERROR;
                     last DO;
                 }
@@ -790,9 +827,10 @@ sub _objDelete
 	        $sql = $delproc;
                 $sql =~ s/\%o/$key/g;
                 $sql =~ s/\%v/$value/g;
+                $sql = $self->_funcParse($sql, $dn);
 
                 # multibyte value
-                $sql = encode($conf->{mbcode}[0], decode('utf8', $sql));
+                $sql = encode($conf->{mbcode}[0], $sql);
 
                 $sth = $self->{db}->prepare("$sql");
                 if (!$sth->execute) {
@@ -809,6 +847,7 @@ sub _objDelete
             $sql = $oconf->{container}[0]->{delproc}[0];
             $sql =~ s/\%o/$key/g;
             $sql =~ s/\%c/$pkey/g;
+            $sql = $self->_funcParse($sql, $dn);
             $sth = $self->{db}->prepare("$sql");
             if (!$sth->execute) {
                 $self->log(level => 'error', message => "Adding link of container by \"$sql\" failed: ".$sth->errstr);
@@ -891,7 +930,7 @@ sub _getParentRdn
 sub _getAttrValues
 {
     my $self = shift;
-    my ($oconf, $attr, $key) = @_;
+    my ($oconf, $key, $pkey, $attr) = @_;
     my $conf = $self->{_config};
     my $aobj = undef;
     my $aoconf = undef;
@@ -906,31 +945,56 @@ sub _getAttrValues
     if (defined($oconf->{attr}{$attr}->{oname})) {
         $aobj = $self->{object}{$oconf->{attr}{$attr}->{oname}[0]};
         $aoconf = $aobj->{conf};
-        $selexpr = "$aoconf->{table}[0].$aoconf->{id}[0]->{column}[0], $aoconf->{table}[0].$aoconf->{attr}{$aoconf->{rdn}[0]}->{column}[0]";
-        $from = "$from, $aoconf->{table}[0]";
-        if (defined($oconf->{attr}{$attr}->{cname})) {
-            $cobj = $self->{object}{$oconf->{attr}{$attr}->{cname}[0]};
+        $selexpr = "$aoconf->{table}[0].$aoconf->{id}[0]->{column}[0]";
+        if (defined($aoconf->{attr}{$aoconf->{rdn}[0]}->{selexpr})) {
+            $selexpr = "$selexpr, $aoconf->{attr}{$aoconf->{rdn}[0]}->{selexpr}[0]";
+        } else {
+            $selexpr = "$selexpr, $aoconf->{table}[0].$aoconf->{attr}{$aoconf->{rdn}[0]}->{column}[0]";
+        }
+
+        if ($oconf->{table}[0] ne $aoconf->{table}[0]) {
+            $from = "$from, $aoconf->{table}[0]";
+        }
+
+        if (defined($aoconf->{parentrel}) &&
+            $aoconf->{parentrel}[0] eq 'tight') {
+            $cobj = $self->{object}{$aoconf->{container}->{oname}[0]};
             $coconf = $cobj->{conf};
-            $selexpr = "$selexpr, $coconf->{table}[0].$coconf->{id}[0]->{column}[0], $coconf->{table}[0].$coconf->{attr}{$coconf->{rdn}[0]}->{column}[0]";
+            $selexpr = "$selexpr, $coconf->{table}[0].$coconf->{id}[0]->{column}[0]";
+            if (defined($aoconf->{attr}{$aoconf->{rdn}[0]}->{selexpr})) {
+                $selexpr = "$selexpr, $coconf->{attr}{$coconf->{rdn}[0]}->{selexpr}[0]";
+            } else {
+                $selexpr = "$selexpr, $coconf->{table}[0].$coconf->{attr}{$coconf->{rdn}[0]}->{column}[0]";
+            }
+
             $from = "$from, $coconf->{table}[0]";
         }
-    } else {
+    } elsif (defined($oconf->{attr}{$attr}->{selexpr})){
         $selexpr = $oconf->{attr}{$attr}->{selexpr}[0];
+    } else {
+        return '';
     }
 
     if (defined($oconf->{attr}{$attr}->{fromtbls})) {
         $from = "$from,$oconf->{attr}{$attr}->{fromtbls}[0]";
     }
 
-    $where = "$oconf->{table}[0].$oconf->{id}[0]->{column}[0] = \'$key\'";
-    if (defined($oconf->{attr}{$attr}->{joinwhere})) {
-        $where = "$where and $oconf->{attr}{$attr}->{joinwhere}[0]";
+    if (defined($oconf->{attr}{$attr}->{where})) {
+        $where = $oconf->{attr}{$attr}->{where}[0];
+        $where =~ s/\%o/$key/g;
+        $where =~ s/\%p/$pkey/g;
+    } else {
+        $where = "$oconf->{table}[0].$oconf->{id}[0]->{column}[0] = \'$key\'";
+        if (defined($oconf->{attr}{$attr}->{joinwhere})) {
+            $where = "$where and $oconf->{attr}{$attr}->{joinwhere}[0]";
+            $where =~ s/\%p/$pkey/g;
+        }
     }
 
     my $sql = "select $selexpr from $from where $where";
     my $sth = $self->{db}->prepare($sql);
     if (!$sth->execute) {
-        $self->log(level => 'error', message => "Getting values by \"$sql\" failed: ".$sth->errstr);
+        $self->log(level => 'error', message => "Getting $attr values by \"$sql\" failed: ".$sth->errstr);
         return undef;
     }
 
@@ -949,7 +1013,12 @@ sub _getAttrValues
         } else{
             $data[0] =~ s/ *$//;
             if ($data[0]) {
-                $attrStr = $attrStr."$attr: $data[0]\n";
+                if ($data[0] =~ /\n/) {
+                    $data[0] = encode_base64($data[0], '');
+                    $attrStr = $attrStr.$attr.":: $data[0]\n";
+                } else {
+                    $attrStr = $attrStr."$attr: $data[0]\n";
+                }
             }
         }
     }
@@ -961,7 +1030,7 @@ sub _getAttrValues
 sub _addAttrValues
 {
     my $self = shift;
-    my ($oconf, $key, $attr, $value) = @_;
+    my ($oconf, $key, $pkey, $attr, $value, $dn, $entryStr) = @_;
     my $conf = $self->{_config};
 
     if (!defined($oconf->{attr}{$attr}->{addproc})) {
@@ -971,6 +1040,7 @@ sub _addAttrValues
     foreach my $addproc (@{$oconf->{attr}{$attr}->{addproc}}) {
         my $sql = $addproc;
         $sql =~ s/\%o/$key/g;
+        $sql =~ s/\%p/$pkey/g;
         if (defined($oconf->{attr}{$attr}->{oname})) {
             my ($rc, $aobj, $attrpkey) = $self->_getObject($value);
             if ($rc) {
@@ -988,7 +1058,8 @@ sub _addAttrValues
             }
             $sql =~ s/\%a/$attrkey/g;
 
-            if (defined($oconf->{attr}{$attr}->{cname})) {
+            if (defined($aobj->{conf}->{parentrel}) &&
+                $aobj->{conf}->{parentrel}[0] eq 'tight') {
                 my $contdn;
                 ($contdn = $value) =~ s/^[^,]+,//;
                 my ($rc, $cobj, $contpkey) = $self->_getObject($contdn);
@@ -1011,13 +1082,15 @@ sub _addAttrValues
             $sql =~ s/\%a/$value/g;
         }
 
+        $sql = $self->_funcParse($sql, $dn, $entryStr);
+
         # multibyte value
-        $sql = encode($conf->{mbcode}[0], decode('utf8', $sql));
+        $sql = encode($conf->{mbcode}[0], $sql);
 
         # add the values to the attribute's table
 	my $sth = $self->{db}->prepare($sql);
         if (!$sth->execute) {
-            $self->log(level => 'error', message => "Adding values by \"$sql\" failed: ".$sth->errstr);
+            $self->log(level => 'error', message => "Adding $attr values by \"$sql\" failed: ".$sth->errstr);
             return -1;
         }
         $sth->finish;
@@ -1029,7 +1102,7 @@ sub _addAttrValues
 sub _delAttrValues
 {
     my $self = shift;
-    my ($oconf, $key, $attr, $value) = @_;
+    my ($oconf, $key, $pkey, $attr, $value, $dn) = @_;
     my $conf = $self->{_config};
 
     if (!defined($oconf->{attr}{$attr}->{delproc})) {
@@ -1039,6 +1112,7 @@ sub _delAttrValues
     foreach my $delproc (@{$oconf->{attr}{$attr}->{delproc}}) {
         my $sql = $delproc;
         $sql =~ s/\%o/$key/g;
+        $sql =~ s/\%p/$pkey/g;
         if ($value) {
             if (defined($oconf->{attr}{$attr}->{oname})) {
                 my ($rc, $aobj, $attrpkey) = $self->_getObject($value);
@@ -1057,7 +1131,8 @@ sub _delAttrValues
                 }
                 $sql =~ s/\%a/$attrkey/g;
 
-                if (defined($oconf->{attr}{$attr}->{cname})) {
+                if (defined($aobj->{conf}->{parentrel}) &&
+                    $aobj->{conf}->{parentrel}[0] eq 'tight') {
                     my $contdn;
                     ($contdn = $value) =~ s/^[^,]+,//;
                     my ($rc, $cobj, $contpkey) = $self->_getObject($contdn);
@@ -1084,13 +1159,15 @@ sub _delAttrValues
             $sql =~ s/\w*\s*=\s*'?\%c'?/1 = 1/g;
         }
 
+        $sql = $self->_funcParse($sql, $dn);
+
         # multibyte value
-        $sql = encode($conf->{mbcode}[0], decode('utf8', $sql));
+        $sql = encode($conf->{mbcode}[0], $sql);
 
         # delete the values from the attribute's table
         my $sth = $self->{db}->prepare($sql);
         if (!$sth->execute) {
-            $self->log(level => 'error', message => "Deleting values by \"$sql\" failed: ".$sth->errstr);
+            $self->log(level => 'error', message => "Deleting $attr values by \"$sql\" failed: ".$sth->errstr);
             return -1;
         }
         $sth->finish;
@@ -1108,7 +1185,7 @@ sub _filter2sql
 
     if (!$filter) {
         ${$from} = $oconf->{table}[0];
-        ${$where} = 1;
+        ${$where} = '1 = 1';
         return 1;
     }
 
@@ -1165,6 +1242,9 @@ sub _filter2sql
         }
     }
 
+    # decode value
+    $value = decode('utf8', $value);
+
     # escape special character
     $value =~ s/'/''/g;
     $value =~ s/\\/\\\\/g;
@@ -1185,6 +1265,7 @@ sub _filter2sql
         ${$where} = "lower($oconf->{table}[0].$oconf->{attr}{$attr}->{column}[0]) $type lower(\'$value\')";
     } elsif (defined($oconf->{attr}{$attr}->{oname})) {
         my $attrobj = $conf->{object}{$oconf->{attr}{$attr}->{oname}[0]};
+        my $aconf = $attrobj->{conf};
         if (defined($oconf->{attr}{$attr}->{fromtbls})) {
             ${$from} = "${$from},$oconf->{attr}{$attr}->{fromtbls}[0]";
         }
@@ -1194,6 +1275,19 @@ sub _filter2sql
         if (defined($oconf->{attr}{$attr}->{joinwhere})) {
             ${$where} = "${$where} and $oconf->{attr}{$attr}->{joinwhere}[0]";
         }
+
+        if (defined($aconf->{container}) && !defined($aconf->{container}[0]->{rdn})) {
+            my $paconf = $self->{object}{$aconf->{container}[0]->{oname}[0]}->{conf};
+
+            $from = "$from, $paconf->{table}[0]";
+            if (defined($aconf->{container}[0]->{fromtbls})) {
+                $from = "$from, $oconf->{container}[0]->{fromtbls}[0]";
+            }
+
+            if (defined($aconf->{container}[0]->{joinwhere})) {
+                $where = "$where and $aconf->{container}[0]->{joinwhere}[0]";
+            }
+        }
     } else {
         if (defined($oconf->{attr}{$attr}->{fromtbls})) {
             ${$from} = "${$from},$oconf->{attr}{$attr}->{fromtbls}[0]";
